#!/usr/bin/env python

""" Integrated Uncertainty and Sensitivity Analysis (iUSA) version 1.
    Performs Variance-Based Global SA of Ideal Point Spatial MCE.

    First order and total order indices are estimated according to the rule in 
    Saltelli A, P Annoni, I Azzini, F Campolongo, M Ratto, S Tarantola, (2010)
    Variance based sensitivity analysis of model output. Design and estimator for the total sensitivity index
    Computer Physics Communications, 181, 259-270

    Total cost = N(k+2) where N is the generated no of samples, k number of model factors (criteria weights)
    BASED ON:   http://sensitivity-analysis.jrc.ec.europa.eu/software/first_total_seq.rar
                Author: Stefano Tarantola
                Joint Research Centre of the European Commission
                Release: 27 September, 2010

    Developer: Arika Ligmann-Zielinska
    Contact address: Department of Geography, 673 Auditorium Rd, Room 121
    Michigan State University East Lansing, MI, 48824-1117
    Email: ligmannz@msu.edu
    
NOTES (last updated July 9, 2014):
*****************************************
[1] The script requires raster files in ASCII grid format (one layer per criterion) and SimLab *.sam file with vectors of weight samples
[2] Only WEIGHTs are factors, crieria surfaces are constant
[3] *.sam file column(factor/weight) order must be the same as the order of input rasters
    e.g. if "distance" is the 2nd raster in the input list then its weighs are taken from the 2nd column in the .sam file
[4] FOR larger rasters MemoryError is raised
    solution - either split input maps into smaller areas and run for each area separately
    or use supercomputing
[5] Uses Sobol quasi-random sampling
    
    Note to self: 
    [1] the format of V arrays: [N(runs), map_cells(cell no), k(factor no)]
    [2] the format of S arrays: [k,map_cells]

"""

#------------- IMPORTS
import sys,os,time
import numpy as npy

# custom imports 
import src.ASCII as ASCII
import src.sample as sample
import src.drule as drule


import Tkinter
import tkFileDialog
#--------------------
def converttime(t):
    """converts time (t) in sec to hm"""
    hours = int(t/3600.0)
    minutes = int((t/3600.0-hours)*60)
    timestr=str(hours)+"h "+str(minutes)+"min "
    return timestr

#------------- INPUTS

inrasters = "./data/ascii/dist.asc;./data/ascii/fcov.asc;./data/ascii/soil.asc"
sobsamfile = None
#------------- APP
try: 
    root = Tkinter.Tk()
    root.withdraw()
    sobsamfile = tkFileDialog.askopenfile(mode='r', filetypes=[("sample file", ".sam")])
    # -------------------------------------------------------
    start = "\n***** "+str(time.strftime("%A, %c")+"\n")
    print start
    t_ini = time.time()
    # INIT -------
    print "SAM (sample) files (Sobol Design):",os.path.basename(sobsamfile.name)
    print "-----"
    # split raster data
    rasterlist = inrasters.split(";")
    # count input rasters
    if len(rasterlist) < 2:
        print "At least two input criteria are required."
        sys.exit(1)
    npy.set_printoptions(threshold=3)
    # SAMPLE FILEs ---------
    k = len(rasterlist)
    sam_dataA = None; sam_dataB = None
    samplesA = None; samplesB = None
    runs = None
    all_sob_sam_data = sample.load_Sobol_SAM_file(sobsamfile.name)
    samplesA = all_sob_sam_data[0][1]; samplesB = all_sob_sam_data[1][1]
    runs = all_sob_sam_data[0][0]
    # generate radial samples
    weightvectors = [block for block in sample.radials(samplesA,samplesB,k)]
    print "Weight samples generated."
    # ASCII GRIDs ---------
    rasters = [];headers = []
    for rasterfile in rasterlist:
        raster = ASCII.Raster(rasterfile)
        rasters.append(raster.body);headers.append(raster.header_str)
        # standardized?
        if npy.amin(raster.body) < 0.0 or npy.amax(raster.body) > 1.0:
            print "\"",raster.name,"\" is not standardized to [0.0, 1.0] scale"
            sys.exit(1)
    # check headers
    if len(set(headers))>1:
        print "ASCII GRIDs have different headers."
        sys.exit(1)
    print len(rasters),"rasters loaded."
    # EXECUTE IDEAL POINT MONTE CARLO for radial blocks ---------
    print "\n**** MODEL EXECUTION"
    map_blocks = [maps for maps in drule.ideal_point_radial(runs,weightvectors,rasters)]
    now = time.time()
    elapsed = now-t_ini
    print "Elapsed time:",converttime(elapsed)
    # VARIANCES & SENSITIVITIES (for 1D maps) -------------------
    # calculation of total variance (Vtot)
    #print "\nBlocks (runs_N_num,factors_k_num,map_cells_size):",npy.array(map_blocks).shape
    y4var =  [IPmap.pop(0) for IPmap in map_blocks] # all A-maps
    y4var += [IPmap.pop(0) for IPmap in map_blocks] # all B-maps
    for amap in y4var:
        amin = amap.min()
        amax = amap.max()
        if amin < 0 or amax > 1:
            "error"
    ay4var = npy.array(y4var)
    Vtot = ay4var.var(axis=0)
    # calculation of nominators for first and totals (Vi,VT)
    map_size = rasters[0].size
    Vi = npy.empty((runs,map_size,k),float)
    VT = npy.empty((runs,map_size,k),float)
    yA_maps = y4var[:runs]
    yB_maps = y4var[runs:]
    mi=npy.mean(Vi[:,:,0])
    for n in range(runs):
        yAb = map_blocks[n]
        yA = yA_maps[n]; yB = yB_maps[n]
        for j in range(k):
            Vi[n,:,j] = yB*(yAb[j]-yA)
            VT[n,:,j] = (yA-yAb[j])**2
    # calculation of sensitivity indices (S,ST)
    S  = npy.empty((k,map_size),float)
    ST = npy.empty((k,map_size),float)
    for j in range(k):
        S[j,:] = npy.mean(Vi[:,:,j],axis=0,dtype=npy.float64)/Vtot
        # GOSH THIS WAS HARD!!! so why this mean formula?
        # get j-th maps for N runs - 2D array of size (N,maps_size)
        # if axis=0 map_cells are preserved, we calculate mean for each cell by collapsing N
        ST[j,:]= npy.mean(VT[:,:,j],axis=0,dtype=npy.float64)/2/Vtot
    # round the values
    S =npy.around(S,decimals=2)
    ST=npy.around(ST,decimals=2)
    # metadata on results
    if npy.any(npy.isnan(S)) or npy.any(npy.isinf(S)):
        print "S maps have NODATA values"
    if npy.any(npy.isnan(ST)) or npy.any(npy.isinf(ST)):
        print "ST maps have NODATA values"
    print "\n\nS min max",S.min(),S.max()
    errors = float(npy.sum(npy.where(S>1,1,0)))
    print "INEXACT S values for max: cells with val of >  1.00 over all factors (k):",int(errors),
    print "[",round(errors/(k*map_size)*100,1),"] %"
    errors = float(npy.sum(npy.where(S<-0.05,1,0)))
    print "INEXACT S values for min: cells with val of < -0.05 over all factors (k):",int(errors),
    print "[",round(errors/(k*map_size)*100,1),"] %"
    print "\n\n"
    print "ST min max",ST.min(),ST.max()
    errors = float(npy.sum(npy.where(ST>2,1,0)))
    print "INEXACT ST values for max: cells with val of >  2.00 over all factors (k):",int(errors),
    print "[",round(errors/(k*map_size)*100,1),"] %"
    errors = float(npy.sum(npy.where(ST<0,1,0)))
    print "INEXACT ST values for min: cells with val of <  0.00 over all factors (k):",int(errors),
    print "[",round(errors/(k*map_size)*100,1),"] %"
    # results dir
    datetime_stamp = 'D%4d_%02d_%02dT%02d_%02d_%02d' % time.localtime()[:6]
    resultdir = os.getcwd()+"/results_"+datetime_stamp+"/"
    os.mkdir(resultdir)
    # save ---> UNCERTAINTY SURFACES
    Amaps = npy.array(yA_maps)
    avg_map = npy.mean(Amaps,axis=0)
    min_map = npy.amin(Amaps,axis=0)
    max_map = npy.amax(Amaps,axis=0)
    std_map = npy.std(Amaps,axis=0)
    rstr = ASCII.Raster(rasterlist[0])# grab one raster for header data
    rstr.save(avg_map,resultdir+"UA_avg_map.asc")
    rstr.save(min_map,resultdir+"UA_min_map.asc")
    rstr.save(max_map,resultdir+"UA_max_map.asc")
    rstr.save(std_map,resultdir+"UA_std_map.asc")
    # save ---> SENSITIVITY SURFACES
    for i,rasterfile in enumerate(rasterlist):
        factor_raster = ASCII.Raster(rasterfile)
        S_for_factor = S[i]
        name = resultdir+"S_"+factor_raster.name+".asc"
        factor_raster.save(S_for_factor,name)
        ST_for_factor = ST[i]
        name = resultdir+"ST_"+factor_raster.name+".asc"
        factor_raster.save(ST_for_factor,name)
    now = time.time()
    elapsed = now-t_ini
    print "Elapsed time:",converttime(elapsed)
    end = "\n*****"+str(time.strftime("%A, %c")+"\n")
    print end
    end = raw_input("Simulations ended. Press any key to close the app...")
except:
     print "\n\n***** ERROR:", sys.exc_info()[0]
     raise
