#!/usr/bin/env python

""" Handles samples for Monte Carlo simulations
    AUTHOR:         Arika Ligmann-Zielinska
    UPDATED:        Apr 27, 2012 
"""
#-------------
import sys, copy

def load_Rand_SAM_file(samfile, factor_number=None):
    """ in: SimLab SAM file (using Random design), number of factors
        out: number of samples & sample list (tuple)
    """
    try:
        f = open(samfile, 'r')
        samrawdata = f.readlines();f.close()
        samrawdata = samrawdata[1:]
        runs = int( samrawdata.pop(0).strip() )
        factors_count = int(samrawdata.pop(0).strip())
        if factor_number:
            if factors_count != factor_number:
                print "ERROR: Number of factors (cols) in the \'",samfile,"\' does not equal number of inputs."
                sys.exit(1)
        samrawdata= samrawdata[1:runs+1]
        samples = [map(float,line.strip().split()) for line in samrawdata]
        return (runs,samples)
    except IOError:
        print "Incorrect format of SimLab *.sam file"

def load_Sobol_SAM_file(samfile):
    """ in: SimLab SAM file (using Sobol design), number of factors
        out: 2 separate -> number of samples & sample list (tuple of 2 tuples)
    """
    #See: Saltelli A, P Annoni, et al. (2010), Variance based sensitivity analysis of model output.
    # Design and estimator for the total sensitivity index, Computer Physics Communications, 181, 259-270
    # specifically p. 263
    try:
        f = open(samfile, 'r')
        samrawdata = f.readlines();f.close()
        samrawdata = samrawdata[1:]
        runs = int( samrawdata.pop(0).strip() )
        factors_count = int(samrawdata.pop(0).strip())
        if factors_count%2: # i.e. is odd not even
            print "ERROR: for Sobol design we need an even number of factors in the *.SAM file"
            sys.exit(1)
        samplesA = None;samplesB = None
        half = factors_count/2
        samrawdata= samrawdata[1:runs+1]
        total = [map(float,rec.strip().split()) for rec in samrawdata]
        samplesA = [line[:half] for line in total]
        samplesB = [line[half:] for line in total]
        return ( (runs,samplesA), (runs,samplesB) )
    except IOError:
        print "Incorrect format of SimLab *.sam file"

def radials(samplesA,samplesB,k):
    """ in: samples A (list of lists), samples B (list of lists), number of factors (k)
        out [generator]: radial samples [A,B,k*Ab]
    """
    runs = len(samplesA)
    simNo = runs*(k+2)
    print "RADIAL MONTE CARLO:",runs,"samples,",k,"factors, &",simNo,"total model runs."
    n = 0
    while n < runs:
        A = samplesA[n];B=samplesB[n]
        sampleblock = [A,B]
        for j in range(k):
            Ab = copy.deepcopy(A)
            Ab[j] = B[j]
            sampleblock.append(Ab)
        yield sampleblock # one radial sample [A,B,k*Ab]
        n += 1
        

#-------------
if __name__ == "__main__":
    sam_a = "./codetestdata/randA128.sam"
    sam_b = "./codetestdata/randB128.sam"
    sam_data_A = load_Rand_SAM_file(sam_a)
    sam_data_B = load_Rand_SAM_file(sam_b)
    #k = 4;sam_data_B = load_SAM_file(sam_b,k) # will generate error
    print "**** RANDOM"
    print "Samples A loaded. total:",sam_data_A[0],"first",sam_data_A[1][0],"last",sam_data_A[1][-1]
    print "Samples B loaded. total:",sam_data_B[0],"first",sam_data_B[1][0],"last",sam_data_B[1][-1]
    k = 3
    vectors = [block for block in radials(sam_data_A[1],sam_data_B[1],k)]
    print "number of radials:",len(vectors)
    print "radial for factors in 'first'"
    for v in vectors[0]: print v
    print "radial for factors in 'last'"
    for v in vectors[-1]: print v
    # Sobol
    sob_sam = "./codetestdata/sob448.sam"
    sob_data = load_Sobol_SAM_file(sob_sam)
    print "**** SOBOL"
    sob_dataA = sob_data[0]; sob_dataB = sob_data[1]
    print "Samples A generated. total:",sob_dataA[0],"first",sob_dataA[1][0],"last",sob_dataA[1][-1]
    print "Samples B generated. total:",sob_dataB[0],"first",sob_dataB[1][0],"last",sob_dataB[1][-1]
    k = 3
    vectors = [block for block in radials(sob_dataA[1],sob_dataB[1],k)]
    print "number of radials:",len(vectors)
    print "radial for factors in 'first'"
    for v in vectors[0]: print v
    print "radial for factors in 'last'"
    for v in vectors[-1]: print v
